% Simple Plus describes one world which is much more complex.
% It contains many event-driven models which describe different features of the world.

?-
  G_Models := 16, % from 1 to 16
  start_random_from(3), % make the random deterministic
  G_Sel := 0,
  G_Pause := 0,
  G_One_Step := 0,
  G_FastSpeed := 0,
  G_Last_Speed := 0,
  G_Results := 0,
  G_Max1 := min_integer,
  G_Max2 := min_integer,
  array(active, 16, 0),
  array(circuits, 16, 0),
  array(time_last, 16, 0),
  array(used_states, 16, 0),
  array(all_arrows, 16, 0),
  array(all_traces, 16, 0),
  array(links3, 16*8*8, 0),
  array(states3, 16*8*4, 0),
  array(events3, 16*8*10, max_integer),
  array(results3, 16*8*4, 0), % G_K =< 4
  array(res2, 8*4, 0), % G_K =< 4
  array(active_last2, 16*2, 0),
  array(events_last2, 16*2, 0),
  array(buf_views, 50, 0),
  array(buf_incorrect, 50, 0),
  array(buf_actions, 50, 0),
  array(buf_arrows2, 16*50, 0),
  array(buf_the_trace2, 16*50, 0),
  init_all_models,
  size(screen, _, H),
  G_Zoom_Font := (H >= 1080 -> 1 else 0.8),
  font(G_Zoom_Font*10, G_Zoom_Font*25, "Courier"), 
  (G_Zoom_Font=:=1 -> X := 1250, Y := 1030 else X := 950, Y := 700),
  window(title("World for AI (Simple Plus)"), size(X, Y), pos(100, 10), paint_indirectly).

op( 700, xfx, ':=='). % for vectors
op( 700, xfx, '::='). % for array2

init_all_models:-
  G_K := 4, % possible values in input/output, from 2 to 4
  G_Time := 0,
  G_Incorrect := 0,
  clean_buffers,
  for( Model, 0, G_Models-1),
    init_model(Model),
    fail.
init_all_models.

init_model(Model):- 
  all_arrows(Model) := 3+random(2), % for arrows, from 1 to 4
  all_traces(Model) := 3+random(2), % for trace,  from 1 to 4
  circuits(Model) := 1+random(2), % increase for more complicated models
  init_events(Model, 0, 0, all_arrows(Model)),
  init_events(Model, 0, 4, all_traces(Model)),
  init_Links(Model).

time_func(end) :-
  check_pause,
  find_correct(Action),
  buf_actions(G_Time mod 50):= Action,

  G_Time := G_Time + 1,
  set_window_text,
%                 observation given by the world
  clean_results,
  G_Max1 := min_integer,
  G_Max2 := min_integer,
  calc_observation(0),
  Min1 := min_integer,
  create_probability(Min1, _, _, _, _, Value, calc),
  buf_views(G_Time mod 50):= Value,
%                 first transition
  transition(1),
%                 action done by the agent
  G_Incorrect := 0,
  calc_observation(1),
  create_incorrect,
  (G_Incorrect =:= 2**G_K - 1 -> find_possible),
  buf_incorrect(G_Time mod 50):= G_Incorrect,

  buf_actions(G_Time mod 50):= G_K,
%                 second transition
  transition(-1),
  update_window(_).

transition(Views):-
  for( Model, 0, G_Models-1),
    transition2(Model, Views),
    fail.
transition(Views).

transition2(Model, Views):-
  events_are_true(Model, Events, Views),
  choose_new_state(Model, New, Events),
  (New =\= active(Model) ->
    (Views =:= 1->
      active_last(Model,0) ::= active(Model),
      active_last(Model,1) ::= New,
      events_last(Model,0) ::= links(Model,active(Model),New) /\ Events,
      events_last(Model,1) ::= 0
    else
      (G_Time =\= time_last(Model) ->
        active_last(Model,0) ::= active(Model),
        active_last(Model,1) ::= active(Model),
        events_last(Model,0) ::= 0
      ),
      events_last(Model,1) ::= links(Model,active(Model),New) /\ Events
    ),
    buf_arrows(Model,G_Time mod 50) ::= events_last(Model,0) \/ events_last(Model,1),
    time_last(Model) := G_Time,
    active(Model) := New
  else (Views =:= 1->
    buf_arrows(Model,G_Time mod 50) ::= 0
  )).

choose_new_state(Model, New, Events):-
  New := active(Model),
  for(I, 0, 7),
    links(Model,active(Model),I) /\ Events =\= 0,
    (New =:= active(Model); random(100) < 33 -> New := I),
    fail.
choose_new_state(Model, New, Events).

set_window_text :-
  set_text("World for AI (Simple Plus), step="+G_Time+(G_Pause =:= 1 -> (G_One_Step > 1 -> " (One Transition)" else " (Pause)") else ""), _).

check_pause:-
  (G_Pause =:= 1 ->
    (G_One_Step =:= 1 ->
      G_One_Step := 0
    else
      G_One_Step > 1,
      (G_One_Step =:= 3 ->
        G_One_Step := 2
      else
        G_Time =\= time_last(G_Sel)
      )
    )
  ).

menu_time(init) :-
  menu(normal, action(menu_pause), text("&Pause\tP") ),
  menu(separator),
  menu(normal, action(menu_one_step), text("&One Step\tO") ),
  menu(separator),
  menu(normal, action(menu_one_transition), text("&One Transition\tI") ),
  menu(separator),
  menu(normal, action(menu_slow_motion), text("&Slow Motion\tS") ),
  menu(checked, action(menu_normal_speed), text("&Normal Speed") ),
  menu(normal, action(menu_fast_speed), text("&Fast Speed\tF") ).

menu_show(init) :-
  menu(normal, action(menu_results), text("Results &List\tL") ).

menu_help(init) :-
  menu(normal, action(menu_description), text("&Description") ),
  menu(normal, action(menu_notetions), text("&Notetions") ).

menu_shift(init) :-
  menu(normal, action(sift_left), text("&Left\t<-") ),
  menu(normal, action(sift_right), text("&Right\t->") ).

menu_change_all_models(press) :-
  init_all_models,
  set_window_text,
  update_window(_).

menu_fast_speed(press) :-
  G_FastSpeed:=(G_FastSpeed=:=1 -> 0 else 1),
  change_speed.
menu_normal_speed(press) :-
  (G_FastSpeed=\=0->
    G_FastSpeed:=0,
    change_speed
  ).
menu_slow_motion(press) :-
  G_FastSpeed:=(G_FastSpeed=:= -1 -> 0 else -1),
  change_speed.

change_speed:-
  modify_menu(G_Menu_Time, 8+G_Last_Speed, normal, _ ),
  modify_menu(G_Menu_Time, 8+G_FastSpeed, checked, _ ),
  G_Last_Speed := G_FastSpeed,
  Interval := (G_FastSpeed =:= 1 -> 0.01 else (G_FastSpeed =:= 0 -> 0.5 else 2)),
  kill_timer(_, G_Timer),
  G_Timer := set_timer(_, Interval, time_func).

menu_results(press) :-
  G_Results := 1 - G_Results,
  modify_menu(G_Menu_Show, 1, (G_Results =:= 1 -> checked else normal), _ ),
  update_window(_).

menu_pause(press) :-
  G_Pause := 1 - G_Pause,
  modify_menu(G_Menu_Time, 1, (G_Pause =:= 1 -> checked else normal), _ ),
  G_One_Step := 0,
  set_window_text.

menu_one_step(press) :-
  (G_Pause =:= 1 ->
    G_One_Step := 1
  ),
  time_func(end).

menu_one_transition(press) :-
  G_Pause := 1,
  modify_menu(G_Menu_Time, 1, checked, _ ),
  G_One_Step := 3,
  time_func(end).

menu_description(press):-
  message("World for AI (Simple Plus)",

  "This world is much more complex. It contains many event-driven models which describe different features of the world.

Every event-driven model has its current state. What is the observation and what actions are correct will be calculated on the base of all event-driven models which work simultaneously.

If different models give different clues then collisions can arise. Look at \"Show\" to see how the observation is calculated.",i).

menu_notetions(press):-
  message("Simple Plus - Notetions",

  "a0=0 means that the action at the current step is zero.

v0=0 means that the observation at the current step is zero.

a1 means the action at the previous step (the current minus one).

a0=\\=0 means that the action at the current step is not zero.

a0*0 means that the action \"zero\" at the current step is correct.

a0#0 means that the action \"zero\" at the current step is not correct.",i).

sift_left(press):-
  move_model(-1).
sift_right(press):-
  move_model(1).

move_model(Direction):-
  G_Sel := G_Sel + Direction,
  (G_Sel < 0 -> G_Sel := 0),
  (G_Sel > G_Models-1 -> G_Sel := G_Models-1),
  update_window(_).

win_func(init) :-
  menu(normal, action(menu_change_all_models), text("&Change all models") ),
  G_Menu_Time := menu( pop_up, action(menu_time), text("&Time") ),
  G_Menu_Show := menu( pop_up, action(menu_show), text("&Show") ),
  menu( pop_up, action(menu_shift), text("&Shift") ),
  menu( right, pop_up, action(menu_help), text("&Help") ),
  G_Timer := set_timer(_, 0.5, time_func).

win_func(size(X,Y)):-
  G_Zoom := Y/967,
  update_window(_).

win_func(key_down(37, Rep)):- % left
  sift_left(press).
win_func(key_down(33, Rep)):- % left (page up)
  sift_left(press).
win_func(key_down(46, Rep)):- % left (delete)
  sift_left(press).
win_func(key_down(35, Rep)):- % left (end)
  sift_left(press).
win_func(key_down(39, Rep)):- % right
  sift_right(press).
win_func(key_down(34, Rep)):- % right (page down)
  sift_right(press).
win_func(key_down(80, Rep)):- % P
  menu_pause(press).
win_func(key_down(79, Rep)):- % O
  menu_one_step(press).
win_func(key_down(73, Rep)):- % I
  menu_one_transition(press).
win_func(key_down(83, Rep)):- % S
  menu_slow_motion(press). 
win_func(key_down(70, Rep)):- % F
  menu_fast_speed(press). 
win_func(key_down(76, Rep)):- % L
  menu_results(press). 

win_func(mouse_click(X, Y)):-
  (G_Zoom_Font=:=1 -> N := 30, X2 := X else N := 24, X2 := X + 36),
  (Y<66, Y>38, abs((X2+N//4) mod N - N//2) =< N//4 ->
    G_Sel := X2 // N - 2
  ),
  update_window(_).

win_func(paint):-
  color_text(_, rgb(0,0,0)),
  text_out(" ", pos(0,10)),
  (G_Zoom_Font=:=1 -> text_out("   ")),
  for(I, 0, G_Models-1),
    text_out(hex_string(I)+" "),
    fail.
win_func(paint):-
  color_text(_, rgb(0,0,255)),
  text_out(" ", pos(0,40)),
  (G_Zoom_Font=:=1 -> text_out("   ")),
  for(I, 0, G_Models-1),
    text_out(active(I)+" "),
    fail.
win_func(paint):-
  pen(2, rgb(255, 0, 0)),
  (G_Zoom_Font=:=1 -> N := 4+2*G_Sel, S := 15 else N := 1+2*G_Sel, S := 12),
  line(N*S-2, 39, (N+1)*S+1, 39, (N+1)*S+1, 65, N*S-2, 65, N*S-2, 39),
  fail.
  
win_func(paint):-
  show_line(0, "views    ", 0, buf_views).
win_func(paint):-
  show_line(0, "incorrect", 40, buf_incorrect).
win_func(paint):-
  show_line(0, "actions  ", 80, buf_actions).
win_func(paint):-
  show_line(G_Sel, "arrows   ", 120, buf_arrows2).
win_func(paint):-
  show_line(G_Sel, "the trace", 160, buf_the_trace2).

win_func(paint):-
  G_Results =:= 0,
  for(Event, 0, all_arrows(G_Sel)-1),
    makeColor(Event, _, Color),
	(2**Event /\ buf_arrows(G_Sel,G_Time mod 50) =\= 0 ->
      pen(10, Color)
    else
      pen(3, Color)
    ),
    line(G_Zoom*600, G_Zoom*(50+Event*70), G_Zoom*1000, G_Zoom*(50+Event*70)),
    line(G_Zoom*970, G_Zoom*(60+Event*70), G_Zoom*1000, G_Zoom*(50+Event*70), G_Zoom*970, G_Zoom*(40+Event*70)),
    pen(3, Color),
    color_text(_, rgb(255,0,0)),
    Star := (event_is_true(G_Sel, Event, 0, 0)-> "*" else " "),
    text_out(Star+2**Event+". ", pos(G_Zoom*600, G_Zoom*(70+Event*70))),
    color_text(_, rgb(0,0,0)),
    print_event(Event, 0),
    for(I, 0, 7),
	  for(J, 0, 7),
	    links(G_Sel,I,J) /\ 2**Event =\= 0,
		R := (active(G_Sel)=:=J -> 40 else 25),
		arrow(I, J, R, Event),
		fail.

win_func(paint):-
  G_Results =:= 0,
  color_text(_, rgb(255,0,0)),
  text_out("The trace", pos(G_Zoom*600, G_Zoom*390)),
  for(Event, 0, all_traces(G_Sel)-1),
    Cell := events(G_Sel,Event+4,0),
    decode_cell(Cell, _, _, Type, _),
    Delay := true_value(Type =:= 2),
    (event_is_true(G_Sel, Event+4, 1, Delay)->
      State := (Delay =:= 1; G_Time =\= time_last(G_Sel) -> active(G_Sel) else active_last(G_Sel,1)),
      Trace := states(G_Sel,State,Event),
      Star := "*"
    else
      Trace := 5,
      Star := " "
    ),
    makeColor(_, Trace, Color),
    pen(10, Color),
    line(G_Zoom*600, G_Zoom*(430+Event*70), G_Zoom*1000, G_Zoom*(430+Event*70)),
    color_text(_, rgb(255,0,0)),
    text_out(Star+2**Event+". ", pos(G_Zoom*600,G_Zoom*(450+Event*70))),
    color_text(_, rgb(0,0,0)),
    print_event(Event+4, 0),
    fail.

win_func(paint):-
  G_Results =:= 1,
  color_text(_, rgb(0,0,0)),
  text_out("everytime", pos(G_Zoom*80,G_Zoom*90)),
  text_out(" never", pos(G_Zoom*250,G_Zoom*90)),
  text_out(" higher", pos(G_Zoom*450,G_Zoom*90)),
  text_out(" lower", pos(G_Zoom*600,G_Zoom*90)),
  color_text(_, rgb(255,0,0)),
  text_out("collisions", pos(G_Zoom*750,G_Zoom*90)),
  color_text(_, rgb(0,0,255)),
  text_out("possibility", pos(G_Zoom*950,G_Zoom*90)),
  make_line(1, 100),
  make_line(2, 250),
  make_line(3, 450),
  make_line(4, 600),
  make_table(160, 0),
  make_table(190, 4),
  fail.

win_func(paint):-
  pen(1, rgb(0, 255 ,0)),
  line(G_Zoom*10, G_Zoom*865, G_Zoom_Font*970, G_Zoom*865),
  (G_Results =:= 1 ->
    line(G_Zoom*110, G_Zoom*160, G_Zoom*110, G_Zoom*740),
    line(G_Zoom*460, G_Zoom*160, G_Zoom*460, G_Zoom*740),
    line(G_Zoom*800, G_Zoom*160, G_Zoom*800, G_Zoom*740),
    line(G_Zoom*950, G_Zoom*160, G_Zoom*950, G_Zoom*740)
  ),
  G_Results =:= 0,
  pen(3, rgb(0,0,0)),
  makeColor(_, 5, Color),
  brush(Color),
  for(I, 0, 7),
    states(G_Sel,I,0) > 0,
    calcCoords(X, Y, I),
	(active(G_Sel)=:=I ->
      R := 40,
      (G_Time mod 2 =:= 1 -> pen(6, rgb(0,0,0)))
    else
      R := 25
    ),
    ellipse(X-R, Y-R, X+R, Y+R),
    draw_trace(I, X, Y, R, 0, 15, -15),
    draw_trace(I, X, Y, R, 1, 10, -5),
    draw_trace(I, X, Y, R, 2, 10, 5),
    draw_trace(I, X, Y, R, 3, 15, 15),
	(active(G_Sel)=:=I ->
      text_out(I, pos(X-6, Y-10))
    ),
    pen(3, rgb(0, 0, 0)),
    fail.

make_table(Y, Row1):-
  for(Row, Row1, Row1+G_K-1),
    color_text(_, rgb(255,0,0)),
    (Row1 =:= 0 ->
      Star := (Row =:= buf_views(G_Time mod 50) -> "*" else " "),
      Text := (G_K > 2, Row>0 -> string_from_ASCII(Row-1+first_ASCII("x")) else Row)
    else
      Star := (Row-4 =:= buf_actions(G_Time mod 50) -> "*" else " "),
      Text := (G_K > 2, Row-4>0 -> string_from_ASCII(Row-5+first_ASCII("a")) else Row-4)
    ),
    text_out(Star+Text+". ", pos(G_Zoom*50, G_Zoom*(Y+Row*70))),
    color_text(_, rgb(0,0,0)),
    X := 120, I := 0,
    for(Col, 0, 7),
      (Col mod 3 =:= 2 ->
        color_text(_, rgb(0,0,255)),
        (Row1 =:= 0, res(Row,0) =:= G_Max1 ->
          color_text_back(_, rgb(255,255,0)),
          text_out(" " + res(Row,2*I) + " ", pos(G_Zoom*X, G_Zoom*(Y+Row*70))),
          color_text_back(_, rgb(255,255,255))
        else (Row1 =:= 4, res(Row,0) =< G_Max2 ->
          (res(Row,0) =:= G_Max2 ->
            color_text_back(_, rgb(255,255,0))
          else
            color_text_back(_, rgb(255,128,128))
          ),
          text_out(" " + res(Row,2*I) + " ", pos(G_Zoom*X, G_Zoom*(Y+Row*70))),
          color_text_back(_, rgb(255,255,255))
        else
          text_out(" " + res(Row,2*I) + " ", pos(G_Zoom*X, G_Zoom*(Y+Row*70)))
        )),
        color_text(_, rgb(0,0,0)),
        I := 1
      else (Col =:= 6 ->
        Collision1 := min( res(Row,0) + res(Row,1), res(Row,1) ),
        Collision2 := min( res(Row,2) + res(Row,3), res(Row,3) ),
        (Collision1+Collision2 > 0 ->
          Text := Collision1+"+"+Collision2+"="+(Collision1+Collision2)
        else
          Text := "."
        ),
        color_text(_, rgb(255,0,0)),
        text_out(Text, pos(G_Zoom*X, G_Zoom*(Y+Row*70)))
      else (Col =:= 7 ->
        (Row1 =:= 0, res(Row,0) =:= G_Max1 ->
          color_text(_, rgb(0,0,255)),
          Min1 := min_integer,
          Text := "",
          create_probability(Min1, _, _, _, Row, Text, paint),
          text_out(Text, pos(G_Zoom*X, G_Zoom*(Y+Row*70)))
        else (Row1 =:= 4 ->
          color_text(_, rgb(255,0,0)),
          text_out(2**(Row-4)+". ", pos(G_Zoom*X, G_Zoom*(Y+Row*70))),
          color_text(_, rgb(0,0,255)),
          (res(Row,0) >= G_Max2 ->
            (res(Row,0) > G_Max2 ->
              Text := "one"
            else (res(Row,2) =:= 0 ->
              Text := "one"
            else (res(Row,2) < 0 ->
              Text := "0.25"
            else
              Text := "one"
            )))
          else
            Text := " 0 "
          ),
          (G_Incorrect /\ 2**(Row-4) =\= 0 ->
            color_text_back(_, rgb(255,128,128)),
            text_out(Text),
            color_text_back(_, rgb(255,255,255))
          else
            text_out(Text)
          )
        ))
      else
        ((Col - I) mod 2 =:= 0 ->
          Result0 := res(Row,Col-I) + res(Row,Col-I+1)
        else
          Result0 := res(Row,Col-I)
        ),
        Text := (Result0 =\= 0 -> Result0 else "."),
        Result := results(G_Sel,Row,Col-I),
        (Result =\= 0 -> Text := Text+" ("+Result+")"),
        text_out(Text, pos(G_Zoom*X, G_Zoom*(Y+Row*70)))
      ))),
      X := X + (Col mod 3 > 0 -> 100 else 150),
      fail.
make_table(Y, Row1).

create_incorrect:-
  for( Action, 0, G_K-1),
    (res(Action+4,0) < G_Max2 ->
      G_Incorrect := G_Incorrect \/ 2**Action
    else (res(Action+4,0) =:= G_Max2, res(Action+4,2) < 0 ->
      (random(20) < 5 -> G_Incorrect := G_Incorrect \/ 2**Action)
    )),
    fail.
create_incorrect.

find_possible:-
  for(Action, 0, G_K - 1),
   res(Action+4,0) =:= G_Max2,
   G_Incorrect := 2**G_K - 1 - 2**Action.

find_correct(Action):-
  Action := random(G_K),
  (G_Incorrect /\ 2**Action =\= 0 ->
    find_correct(Action)
  ).

create_probability(Min1, Min2, Number, NumberMin2, Row, Text, _):-
  Number := 0,
  Min2 := max_integer,
  for(I, 0, G_K-1),
    res(I,0) =:= G_Max1,
    res(I,2) > Min1,
    Number := Number + 1,
    (Min2 > res(I,2) ->
      NumberMin2 := 1,
      Min2 := res(I,2)
    else (Min2 =:= res(I,2) ->
      NumberMin2 := NumberMin2 + 1
    )),
    fail.
create_probability(_, Min2, Number, NumberMin2, Row, Text, paint):-
  (Min2 < res(Row,2) ->
    Text := Text + "("+(2*Number-NumberMin2)+"/"+2*Number+")*",
    create_probability(Min2, _, _, _, Row, Text, paint)
  else (Number =\= NumberMin2 ->
    Text := Text + "(1/"+2*Number+")"
  else (Number =\= 1 ->
    Text := Text + "(1/"+Number+")"
  else
    Text := Text + "one"
  ))).
create_probability(_, Min2, Number, NumberMin2, _, Value, calc):-
  (Number =:= NumberMin2 ->
    ItsNumber := random(Number),
    find_its_number(Min2, ItsNumber, Value)
  else
    ItsNumber := random(2*Number),
    (ItsNumber < NumberMin2 ->
      find_its_number(Min2, ItsNumber, Value)
    else
      create_probability(Min2, _, _, _, _, Value, calc)
    )
  ).

find_its_number(Min1, ItsNumber, Value):-
  Number := 0,
  for(I, 0, G_K-1),
    res(I,0) =:= G_Max1,
    res(I,2) =:= Min1,
    Number := Number + 1,
    ItsNumber < Number,
    Value := I.

make_line(Col, X):-
  makeColor(_, Col, Color),
  pen(10, Color),
  line(G_Zoom*X, G_Zoom*130, G_Zoom*(X+130), G_Zoom*130).

draw_trace(I, X, Y, R, Num, X1, Y1):-
  State := states(G_Sel,I,Num),
  makeColor(_, State, Color),
  pen(10, Color),
  line(X-R+X1,Y+Y1,X+R-X1,Y+Y1).

print_event(Event, Count):-
  Cell := events(G_Sel,Event,Count),
  (Cell =\= max_integer ->
    print_event(Event, Count+1),
    decode_cell(Cell, Negative, Value, Type, Atom),
    (Count=:=0 ->
      (Event>3 -> text_out(" => ")),
      color_text(_, rgb(0,0,255))
    ),
    (G_K > 2, Value > 0 ->
      Value := string_from_ASCII(Value-1+first_ASCII((Type =:= 2 -> "x" else "a")))
    ),
    (Type =:= 1 ->
      text_out("a" + Atom + (Negative =:= 0 -> "*" else "#") + Value)
    else
      text_out((Type =:= 2 -> "v" else "a") + Atom + (Negative =:= 0 -> "=" else "=\\=") + Value)
    ),
    (Count>0, not(Count =:= 1 , Event>3) ->
      text_out(" & ")
    )
  ).

event_is_true(Model, Event, Count, Delay):-
  Cell := events(Model,Event,Count),
  (Cell =\= max_integer ->
    decode_cell(Cell, Negative, Value, Type, Atom),
    (Type =:= 1 ->
      Res := true_value(buf_incorrect((G_Time+Delay-Atom) mod 50) /\ 2**Value =:= 0)
    else
      Buffer := (Type =:= 0 -> buf_actions else buf_views),
      Res := true_value(Buffer((G_Time+Delay-Atom) mod 50) =:= Value)
    ),
    (Res + Negative) mod 2 =:= 1,
    event_is_true(Model, Event, Count+1, Delay)
  ).

events_are_true(Model, Events, Views):-
  Events := 0,
  for(Event, 0, all_arrows(Model)-1),
    Cell := events(Model,Event,0),
    (Cell =:= max_integer ->
      (Views =:= -1 -> Events := Events \/ 2**Event)
    else
      decode_cell(Cell, _, _, Type, _),
      (Views*(Type - 1.5) > 0 , event_is_true(Model, Event, 0, 0) -> Events := Events \/ 2**Event)
    ),
    fail.
events_are_true(Model, Events, Views).

show_line(Model, Title, Y, Buffer):-
  color_text(_, rgb(0,0,0)),
  text_out(" "+Title+" -> ", pos(0,G_Zoom*(750+Y))),
  for(N, 1, 50),
    (N mod 5 =:=0 -> color_text(_, rgb(255,0,0))),
    (N mod 5 =:=1 -> color_text(_, rgb(0,0,255))),
    Value := Buffer(50*Model + (N+G_Time) mod 50),
    (Value=:=0 ->
      text_out(".")
    else (Y=\=0, Y=\=80 ->
      text_out(hex_string(Value))
    else (Value=:=G_K ->
      text_out("_")
    else (G_K =:= 2 ->
      text_out("*")
    else (Y=:=0 ->
      text_out(string_from_ASCII(Value-1+first_ASCII("x")))
    else
      text_out(string_from_ASCII(Value-1+first_ASCII("a")))
    ))))),
    fail.

arrow(I, J, R, Event):-
  calcCoords(X1,Y1,I),
  calcCoords(X2,Y2,J),
  D:= sqrt((X1-X2)**2+(Y1-Y2)**2),
  [A,B]:==[X1-X2,Y1-Y2]/D,
  Sign := (X1>X2; X1=X2, Y1<Y2->1 else -1),
  [X1,Y1]:==[X1,Y1]+(Sign*10*(Event-1.5))*[B,-1*A],
  [X2,Y2]:==[X2,Y2]+(Sign*10*(Event-1.5))*[B,-1*A],

  [X3,Y3]:==[X2,Y2]+R*[A,B],
  [X4,Y4]:==[X2,Y2]+(R+30*G_Zoom)*[A,B],
  [X4a,Y4a]:==[X4,Y4]+10*G_Zoom*[-1*B,A],
  [X4b,Y4b]:==[X4,Y4]+10*G_Zoom*[B,-1*A],
  makeColor(Event, _, Color),
  (I =:= active_last(G_Sel,0), J =:= active_last(G_Sel,1), 2**Event /\ events_last(G_Sel,0) =\= 0
  ;
   I =:= active_last(G_Sel,1), J =:= active(G_Sel), 2**Event /\ events_last(G_Sel,1) =\= 0 ->
    pen(10, Color)
  else
    pen(3, Color)
  ),
  line(X4a,Y4a,X3,Y3,X4b,Y4b),
  line(X1,Y1,X3,Y3).

calcCoords(X,Y,I):-
  X := G_Zoom*(300 - 200*cos(I*pi/4)),
  Y := G_Zoom*(400 - 200*sin(I*pi/4)).

makeColor(0, 0, rgb(0,0,0)). % arrow, trace
makeColor(1, 1, rgb(255,0,0)).
makeColor(2, 2, rgb(0,0,255)).
makeColor(3, 4, rgb(0,128,0)).
makeColor(4, 3, rgb(255,255,0)).
makeColor(5, 5, rgb(0,255,255)).

calc_observation(Action):-
  for( Model, 0, G_Models-1),
    calc_observation2(Model, Action).
calc_observation(Action):-
  for(Value, 4*Action, 4*Action+G_K-1),
    for(I, 0, 1),
      New := 0,
      New1 := 0,
      add_results(New, New1, Value, I),
      res(Value,2*I) ::= New,
      res(Value,2*I+1) ::= New1,
      (Action =:= 0 -> 
        (G_Max1 < New, I =:= 0 -> G_Max1 := New)
      else
        (G_Max2 < New, I =:= 0 ->
          G_Max2 := (New < 0 -> New else 0)
        )
      ),
      fail.
calc_observation(Action).

calc_observation2(Model, Action):-
  for(Event, 0, all_traces(Model)-1),
    State := states(Model,active(Model),Event) - 1,
    State < 4,
    Cell := events(Model,Event+4,0),
    decode_cell(Cell, _, Value, Type, _),
    Type - 1 =\= Action,
    Val2 := 4*Action + Value,
    (event_is_true(Model, Event+4, 1, 0)->
      results(Model,Val2,State) ::= results(Model,Val2,State) + 1,
      buf_the_trace(Model,G_Time mod 50) ::= buf_the_trace(Model,G_Time mod 50) \/ 2**Event
    ),
    fail.

add_results(New, New1, Value, I):-
  for( Model, 0, G_Models-1),
    New := New + results(Model,Value,2*I) - results(Model,Value,2*I+1),
    New1 := New1 + results(Model,Value,2*I+1),
    fail.
add_results(New, New1, Value, I).

clean_buffers:- for(I, 0, 16-1), active(I):=0, time_last(I):=0, used_states(I):=1, fail.
clean_buffers:- for(I, 0, 15), for(J, 0, 1), active_last(I,J) ::= 0, events_last(I,J) ::= 0, fail.
clean_buffers:- for(I, 0, 15), for(J, 0, 49), buf_arrows(I,J) ::= 0, buf_the_trace(I,J) ::= 0, fail.
clean_buffers:- for(I, 0, 15), for(J, 0, 7), for(K, 0, 7), links(I,J,K) ::= 0, fail.
clean_buffers:- for(I, 0, 15), for(J, 0, 7), for(K, 0, 3), states(I,J,K) ::= 0, fail.
clean_buffers:- for(I, 0, 15), for(J, 0, 7), for(K, 0, 9), events(I,J,K) ::= max_integer, fail.
clean_buffers:- for(I, 0, 50-1), buf_views(I):=0, buf_incorrect(I):=0, buf_actions(I):=0, fail.
clean_buffers.

clean_results:- for(I, 0, 15), for(J, 0, 7), for(K, 0, 3), results(I,J,K) ::= 0, fail.
clean_results:- for(Model, 0, 16-1), buf_the_trace(Model,G_Time mod 50) ::= 0, fail.
clean_results.

init_events(Model, Event, Start, AllEvents):-
  (Event < AllEvents ->
    init_one_event(Model, Event, Start),
    (not(exist_the_same(Model, Event, Start))->
      Event := Event + 1
    ),
    init_events(Model, Event, Start, AllEvents)
  ).

init_one_event(Model, Event, Start):-
  (Start =:= 0 ->
    random(4)=\=0, % make the event "True"
    First := random(3),
    FirstVal := (G_K =:= 2, First =\= 1 -> 2*random(2) else random(2*G_K))
  else
    First := random(2)+1,
    FirstVal := (G_K =:= 2, First =\= 1 -> 2 else 2*random(G_K))
  ),
  event_tail(Model, Event + Start, First, FirstVal, 0).
init_one_event(Model, Event, Start).

event_tail(Model, Event, Last, LastVal, Count):-
  encode_cell(Cell, LastVal, 0, Last, 0),
  events(Model,Event,Count) ::= Cell,
  Count < 3, random(3)=\=0,
  Next := Last + 1 + random(3),  % increase 3 to look further in the past
  NextVal := (G_K =:= 2, Next mod 3 =\= 1 -> 2*random(2) else random(2*G_K)),  
  event_tail(Model, Event, Next, NextVal, Count+1).
  
exist_the_same(Model, This, Start):-
  for(Event, 0, This - 1),
    the_same_event(Model, This + Start, Event + Start, 0).

the_same_event(Model, E1, E2, Count):-
  events(Model,E1,Count) =:= events(Model,E2,Count),
  (events(Model,E1,Count) =:= max_integer; the_same_event(Model, E1, E2, Count + 1)).

init_Links(Model):-
  G_Used_Arrows := 0,
  G_Used_Traces := 1,
  states(Model,0,0) ::= random(4)+1, % for more trace
  states(Model,0,1) ::= 5, states(Model,0,2) ::= 5, states(Model,0,3) ::= 5, 
  for(I, 1, circuits(Model)),
    circuit(Model, 0).
init_Links(Model):-
  all_arrows(Model) := G_Used_Arrows,
  all_traces(Model) := G_Used_Traces.

circuit(Model, State):-
  State2 := random(7),
  (State=:=State2; State2 >= used_states(Model) ->
    (used_states(Model) < 8 ->
      State2 := used_states(Model),
      used_states(Model) := used_states(Model) + 1
    else
      State2 := 7
    )
  ),
  Event := random(min(G_Used_Arrows+1, all_arrows(Model))),
  (Event =:= G_Used_Arrows ->
    G_Used_Arrows := G_Used_Arrows+1
  ),
  links(Model,State,State2) ::= links(Model,State,State2) \/ 2**Event,
  (states(Model,State,0) =:= 0 ->
    make_trace(Model, State, 0),
    make_trace(Model, State, 1),
    make_trace(Model, State, 2),
    make_trace(Model, State, 3)
  ),
  State2 =\= 0,
  circuit(Model, State2).

make_trace(Model, State, Event):-
  (Event > G_Used_Traces; Event >= all_traces(Model) ->
    Trace := 5
  else
    Trace := min(random(7), 4)+1,
    (Event =:= G_Used_Traces, Trace<5 ->
      G_Used_Traces := G_Used_Traces+1
    )
  ),
  states(Model,State,Event) ::= Trace.

[] :== []/S. 
[H|T] :== [H1|T1]/S :- 
  H:= H1/S,
  T:== T1/S.

[] :== []+S*[]. 
[H|T] :== [H1|T1]+S*[H2|T2] :- 
  H:=H1+S*H2,
  T:==T1+S*T2.

encode_cell(Cell, Negative, Value, Type, Atom):-
  Cell := Negative + 2*(Value + G_K*(Type + 3*Atom)).
decode_cell(Cell, Negative, Value, Type, Atom):-
  Negative := Cell mod 2,
  Value := (Cell//2) mod G_K,
  Type := (Cell//G_K//2) mod 3,
  Atom := Cell//3//G_K//2.

% Type is a[]= or a[]* or v[]=
% Type[Atom] = Value

R is links(int X, int Y, int Z) :- R := links3(64*X+8*Y+Z).
R is states(int X, int Y, int Z) :- R := states3(32*X+4*Y+Z).
R is events(int X, int Y, int Z) :- R := events3(80*X+10*Y+Z).
R is results(int X, int Y, int Z) :- R := results3(32*X+4*Y+Z).
R is res(int X, int Y) :- R := res2(4*X+Y).
R is active_last(int X, int Y) :- R := active_last2(2*X+Y).
R is events_last(int X, int Y) :- R := events_last2(2*X+Y).
R is buf_arrows(int X, int Y) :- R := buf_arrows2(50*X+Y).
R is buf_the_trace(int X, int Y) :- R := buf_the_trace2(50*X+Y).

links(X, Y, Z) ::= Val :- links3(64*X+8*Y+Z) := Val.
states(X, Y, Z) ::= Val :- states3(32*X+4*Y+Z) := Val.
events(X, Y, Z) ::= Val :- events3(80*X+10*Y+Z) := Val.
results(X, Y, Z) ::= Val :- results3(32*X+4*Y+Z) := Val.
res(X, Y) ::= Val :- res2(4*X+Y) := Val.
active_last(X, Y) ::= Val :- active_last2(2*X+Y) := Val.
events_last(X, Y) ::= Val :- events_last2(2*X+Y) := Val.
buf_arrows(X, Y) ::= Val :- buf_arrows2(50*X+Y) := Val.
buf_the_trace(X, Y) ::= Val :- buf_the_trace2(50*X+Y) := Val.
