% AI Unravels Chess
% version 9 from 7 March, 2023
% previous version is at https://dobrev.com/software/AI_unravels_chess_v8.zip

include("AI_init.h").
include("AI_window.h").
include("AI_data.h").
include("AI_our_ai.h").

?-
  G_M := 27, % Models
  G_S := 20, % States (17)
  G_E := 70, % Events
  G_Tr := 14, % Traces
  G_T := 18, % Events+Traces
  G_K := 4, % possible values in input/output, from 2 to 4
  G_Sel := 0,
  G_Sel_Tr_V := -1,
  G_Sel_Tr_H := 1,
  G_Sel_Show := 0,
  G_Pause := 1,
  G_Start := 1,
  G_Agent := 0,
  G_Next_Action := -1,
  G_Repeat_Action := 0,
  G_with_Randomness := 0,
  G_Observed := 0,
  G_ObserveNow := 0,
  G_One_Step := 0,
  G_FastSpeed := 0,
  G_ModStat := 2,
  G_Results := 0,
  G_Max1 := min_integer,
  G_Max2 := min_integer,
  array(enter, G_M, -1),
  array(exit, G_M, -1),
  array(active, G_M, 1),
  array(restricted, G_M, 0),
  array(time_last, G_M, 0),
  array(all_arrows, G_M, 0),
  array(all_traces, G_M, 0),
  array(active_last2, G_M*2, 1),
  array(events_last2, G_M*2, 0),
  array(links3, G_M*G_S*G_S, 0),
  array(states3, G_M*G_S*G_Tr, 0),
  array(events2, G_M*G_T, -1),
  array(event_cell2, G_E*10, max_integer),
  array(events_val, G_E, 0),
  array(results3, G_M*8*4, 0), % G_K =< 4
  array(res2, 8*4, 0), % G_K =< 4
  array(buf_views, 50, 0),
  array(buf_incorrect, 50, 0),
  array(buf_actions, 50, 0),
  array(buf_arrows2, G_M*50, 0),
  array(buf_the_trace2, G_M*50, 0),
  array(board2, 8*8+2, 0),
  our_ai_define,
  init_all,
  size(screen, _, H),
  (H >= 1080 ->
    X := 1270, Y := 1030,
    G_Zoom_Font := 1,
    G_ChS := 15
  else
    X := 1020, Y := 760,
    G_Zoom_Font := 0.8,
    G_ChS := 12
  ),
  font(G_Zoom_Font*10, G_Zoom_Font*25, "Courier"),
  window(title("AI Unravels Chess"), size(X, Y), pos(100, 10), paint_indirectly).

init_all:-
  start_random_from(0), % make the random models deterministic
  G_Time := -1,
  G_Incorrect := 0,
  clean_buffers,
  init_board,
  fixed_events,
  fixed_models,
  clean_results,
  our_ai_init,
  active(0) := 4,

      %active(1) := 2**2,
      %active(2) := 2**5,
      %board(4,5) ::= 5,
      %board(2,5) ::= 4,

  for( Model, 16, 25),
    init_model(Model),
    fail.
init_all:-
  start_random_from(0), % make the life deterministic
  make_one_step,
  buf_actions(0):= G_K.

time_func(end) :-
  check_pause,
  (G_Next_Action =\= -1 ->
    Action := G_Next_Action,
    G_Next_Action := -1
  else (G_Agent =:= 0 -> % Random
    find_correct(Action)
  else (G_Agent =:= 1 -> % Freeze
    Action := 0
  else
    our_ai_action(Action)
  ))),
  (G_Incorrect /\ 2**Action =\= 0 ->
    buf_tried(G_Time mod 50) := buf_tried(G_Time mod 50) \/ 2**Action,
    Action := G_K+Action,
    G_Repeat_Action := 1
  else
    G_Repeat_Action := 0
  ),
  buf_actions(G_Time mod 50):= Action,
  (G_Repeat_Action=:=0 ->
    make_one_step,
    buf_actions(G_Time mod 50):= G_K
  ),
  set_window_text,
  update_window(_).

make_one_step:-
%                 initialization
    clean_results,
    enter_or_exit_algorithm(enter, 1),
    enter_or_exit_algorithm(exit, 0),

    make_restriction(3),
%                 second transition
    evaluete_events(_, transition),
    transition(-1),

%                 zero (fourth) restriction
    make_restriction(4),
    update_memory,
    G_Time := G_Time + 1,
%                 observation given by the world
    G_Max1 := min_integer,
    G_Max2 := min_integer,
    evaluete_events(_, forcast),
    calc_observation(0),
    Min1 := min_integer,
    create_probability(Min1, _, _, _, _, Value, calc),
    buf_views(G_Time mod 50):= Value,
    make_restriction(1),
%                 first transition
    evaluete_events(_, transition),
    transition(1),
    make_restriction(2),
%                 action done by the agent
    G_Incorrect := 0,
    evaluete_events(_, forcast),
    calc_observation(1),
    create_incorrect,
    (G_Incorrect =:= 2**G_K - 1 -> find_possible),
    buf_incorrect(G_Time mod 50):= G_Incorrect,
    our_ai_thinks.

enter_or_exit_algorithm(What, N):-
  for(Model, 0, G_M-1),
  Event2:= What(Model),
  Event2 =\= -1,
  Cell := event_cell(Event2,0),
  decode_cell(Cell, _, _, Type, _),
  Delay := true_value(Type =:= 2),
  (event_is_true(Event2, 0, Delay)->
    active(Model):=N
  ),
  fail.
enter_or_exit_algorithm(What, N).

transition(Views):-
  for( Model, 0, G_M-1),
    transition2(Model, Views),
    fail.
transition(Views).

transition2(Model, Views):-
  events_are_true(Model, Events, Views),
  choose_new_states(Model, Moved, New, Events, New_E),
  (New =\= 0 ->
    New2 := New \/ active(Model) /\ \Moved, 
    (Views =:= -1 ->
      active_last(Model,0) ::= active(Model),
      active_last(Model,1) ::= New2,
      events_last(Model,0) ::= New_E,
      events_last(Model,1) ::= 0
    else
      (G_Time =\= time_last(Model) ->
        active_last(Model,0) ::= active(Model),
        active_last(Model,1) ::= active(Model),
        events_last(Model,0) ::= 0
      ),
      events_last(Model,1) ::= New_E
    ),
    buf_arrows(Model,G_Time mod 50) ::= events_last(Model,0) \/ events_last(Model,1),
    time_last(Model) := G_Time,
    active(Model) := New2
  else (Views =:= 1->
    buf_arrows(Model,G_Time mod 50) ::= 0
  )).

choose_new_states(Model, Moved, New, Events, New_E):-
  Moved := 0,  New := 0, New_E := 0,
  Sn := find_any_1_in(active(Model)),
    for(I, 0, G_S-1),
      E := links(Model,Sn,I) /\ Events,
      E =\= 0,
        Moved := Moved\/2**Sn,
        New := New\/2**I,
        New_E := New_E\/E,
        fail.
choose_new_states(Model, Moved, New, Events, New_E).

create_incorrect:-
  for( Action, 0, G_K-1),
    (res(Action+4,0) < G_Max2 ->
      G_Incorrect := G_Incorrect \/ 2**Action
    else (res(Action+4,0) =:= G_Max2, res(Action+4,2) < 0 ->
      (random(20) < 5 -> G_Incorrect := G_Incorrect \/ 2**Action)
    )),
    fail.
create_incorrect.

find_possible:-
  for(Action, 0, G_K - 1),
   res(Action+4,0) =:= G_Max2,
   G_Incorrect := 2**G_K - 1 - 2**Action.

find_correct(Action):-
  Action := random(G_K),
  (G_Incorrect /\ 2**Action =\= 0 ->
    find_correct(Action)
  ).

create_probability(Min1, Min2, Number, NumberMin2, Row, Text, _):-
  Number := 0,
  Min2 := max_integer,
  for(I, 0, G_K-1),
    res(I,0) =:= G_Max1,
    res(I,2) > Min1,
    Number := Number + 1,
    (Min2 > res(I,2) ->
      NumberMin2 := 1,
      Min2 := res(I,2)
    else (Min2 =:= res(I,2) ->
      NumberMin2 := NumberMin2 + 1
    )),
    fail.
create_probability(_, Min2, Number, NumberMin2, Row, Text, paint):-
  (Min2 < res(Row,2) ->
    Text := Text + "("+(2*Number-NumberMin2)+"/"+2*Number+")*",
    create_probability(Min2, _, _, _, Row, Text, paint)
  else (Number =\= NumberMin2 ->
    Text := Text + "(1/"+2*Number+")"
  else (Number =\= 1 ->
    Text := Text + "(1/"+Number+")"
  else
    Text := Text + "one"
  ))).
create_probability(_, Min2, Number, NumberMin2, _, Value, calc):-
  (Number =:= NumberMin2 ->
    ItsNumber := (G_with_Randomness =:= 0 -> 0 else random(Number)),
    find_its_number(Min2, ItsNumber, Value)
  else
    ItsNumber := (G_with_Randomness =:= 0 -> 2*Number-1 else random(2*Number)),
    (ItsNumber < NumberMin2 ->
      find_its_number(Min2, ItsNumber, Value)
    else
      create_probability(Min2, _, _, _, _, Value, calc)
    )
  ).

find_its_number(Min1, ItsNumber, Value):-
  Number := 0,
  for(I, 0, G_K-1),
    res(I,0) =:= G_Max1,
    res(I,2) =:= Min1,
    Number := Number + 1,
    ItsNumber < Number,
    Value := I.

make_restriction(Views):-
  evaluete_events(Views, restriction),
  for( Model, 0, G_M-1),
    State := find_any_1_in(active(Model)),
      for(Event, 0, all_traces(Model)-1),
        Trace := states(Model,State,Event),
        Trace < 3,
        Event2 := events(Model,Event+4),
        events_val(Event2) > 0,
        (Trace + events_val(Event2)) mod 2=:=1,
        New := active(Model)/\ \2**State,
        (New=\=0->
          restricted(Model):=restricted(Model)\/2**State,
          active(Model):=New
        ),
        fail.
make_restriction(_).

calc_observation(Action):-
  for( Model, 0, G_M-1),
    observable(Model),
    calc_observation2(Model, Action).
calc_observation(Action):-
  for(Value, 4*Action, 4*Action+G_K-1),
    for(I, 0, 1),
      New := 0,
      New1 := 0,
      add_results(New, New1, Value, I),
      res(Value,2*I) ::= New,
      res(Value,2*I+1) ::= New1,
      (Action =:= 0 -> 
        (G_Max1 < New, I =:= 0 -> G_Max1 := New)
      else
        (G_Max2 < New, I =:= 0 ->
          G_Max2 := (New < 0 -> New else 0)
        )
      ),
      fail.
calc_observation(Action).

calc_observation2(Model, Action):-
  for(Event, 0, all_traces(Model)-1),
    Trace := states(Model, find_first_1_in(active(Model)), Event),
    if_all_agree(Model, active(Model), Event, Trace),
    Trace < 5,
    Event2 := events(Model,Event+4),
    decode_events_val(events_val(Event2), Negative, Value, Action2, Flag),
    (Action2 =:= Action+1->
      Add := Trace-1,
      (Flag=:=0 ->
        (Negative=:=1 ->
          Add := (Trace mod 2=:=1 -> Add+1 else Add-1)
        ),
        results(Model,4*Action+Value,Add) ::= results(Model,4*Action+Value,Add) + 1
      else
        (Trace mod 2=:=1 ->
          Add := Add+1
        ),
        ((Negative+Trace) mod 2=:=0 ->
          results(Model,4*Action+Value,Add) ::= results(Model,4*Action+Value,Add) + 1
        else
          add_to_all(Model,4*Action+Value-1,Add),
          results(Model,4*Action+Value,Add) ::= results(Model,4*Action+Value,Add) - 1
        )
      ),
      buf_the_trace(Model,(G_Time-1) mod 50) ::= buf_the_trace(Model,(G_Time-1) mod 50) \/ 2**Event
    ),
    fail.

if_all_agree(Model, States, Event, Trace):-
  State := find_any_1_in(States),
  Trace2 := states(Model,State,Event),
  Trace =\= Trace2,
  (Trace2<3->
    Trace =\= Trace2+2
  else (Trace2<5->
    (Trace =:= Trace2-2 ->
      Trace := Trace2,
      fail
    )
  )),
  !,
  Trace := 5.
if_all_agree(Model, States, Event, Trace).

add_to_all(Model,Value,Add):-
  for(Value, 0, G_K-1),
    results(Model,Value+1,Add) ::= results(Model,Value+1,Add) + 1,
    fail.
add_to_all(Model,Value,Add).

add_results(New, New1, Value, I):-
  for( Model, 0, G_M-1),
    New := New + results(Model,Value,2*I) - results(Model,Value,2*I+1),
    New1 := New1 + results(Model,Value,2*I+1),
    fail.
add_results(New, New1, Value, I).

clean_buffers:- for(I, 0, G_M-1), active(I):=true_value(I<5; 15<I, I=\=26), time_last(I):=0, fail.
clean_buffers:- for(I, 0, G_M*2-1), active_last2(I):=1, events_last2(I):=0, fail.
clean_buffers:- for(I, 0, G_M*G_S*G_S-1), links3(I):=0, fail.
clean_buffers:- for(I, 0, G_M*G_S*G_Tr-1), states3(I):=0, fail.
clean_buffers:- for(I, 0, 50-1), buf_views(I):=0, buf_incorrect(I):=0, buf_actions(I):=0, fail.
clean_buffers:- for(I, 0, G_M*50-1), buf_arrows2(I):=0, buf_the_trace2(I):=0, fail.
clean_buffers:- for(I, 0, G_M*8-1), events2(I):= -1, fail.
clean_buffers:- for(I, 0, G_E*10-1), event_cell2(I):=max_integer, fail.
clean_buffers:- for(I, 0, 66-1), board2(I):=0, fail.
clean_buffers.

clean_results:- for(I, 0, G_M*8*4-1), results3(I):=0, fail.
clean_results:- for(I, 0, 8*4-1), res2(I):=0, fail.
clean_results:-
  buf_tried((G_Time+1) mod 50) := 0,
  for(Model, 0, G_M-2),
    restricted(Model) := 0,
    buf_the_trace(Model,G_Time mod 50) ::= 0,
    fail.
clean_results.

clean_model(Model):-
  for(I, 0, 7),
    for(J, 0, 7),
      links(Model,I,J) ::= 0,
      fail.
clean_model(Model):-
  for(State, 0, 7),
    for(Event, 0, 2),
      states(Model,State,Event) ::= 0,
      fail.
clean_model(Model).

init_model(Model):-
  all_arrows(Model) := 2,
  events(Model,0) ::= 1,
  events(Model,1) ::= 2,

  all_traces(Model) := 3,
  events(Model,4) ::= 36,
  events(Model,5) ::= 37,
  events(Model,6) ::= 38,
  init_Links(Model).

init_Links(Model):-
  Used := 0,
  add_red(Model, Used, 6),
  add_black(Model, Used),
  (cannot_go_somewhere(Model, Used)->
    clean_model(Model),
    init_Links(Model)
  ).

add_red(Model, Used, Cycle):-
  Cycle := min(8-Used, 1+random(Cycle)),
  for(I, 0, Cycle-1),
    make_trace(Model, Used+I),
    J := (I+1) mod Cycle,
    links(Model,Used+I,Used+J) ::= 2**1,
    fail.
add_red(Model, Used, Cycle):-
  Begin := Used,
  Used := Used + Cycle,
  (Used<8, random(3)=:=0 ->
    make_trace(Model, Used),
    links(Model,Used,Used-1) ::= 2**1,
    Used := Used + 1,
    add_one_more(Model, Begin, Used)
  ),
  (Used<5 ->
    add_red(Model, Used, Cycle)
  ).

make_trace(Model, State):-
  for(Event, 0, 2),
    Trace := 2*random(8)+1,
    (Trace>4 -> Trace :=5),
    states(Model,State,Event) ::= Trace,
    fail.
make_trace(Model, State).

add_one_more(Model, Begin, Used):-
  (Used<8, random(3)=:=0 ->
    make_trace(Model, Used),
    links(Model,Used,Begin+random(Used-Begin)) ::= 2**1,
    Used := Used + 1,
    add_one_more(Model, Begin, Used)
  ).

add_black(Model, Used):-
  for(I, 0, Used-1),
    J := random(Used),
    links(Model,I,J) ::= links(Model,I,J) \/ 2**0,
    fail.
add_black(Model, Used).

cannot_go_somewhere(Model, Used):-
  for(I, 0, Used-1),
    not(can_go(Model, I, 0, 2**I)).
cannot_go_somewhere(Model, Used):-
  for(I, 0, Used-1),
    not(can_go(Model, 0, I, 1)).

can_go(Model, I, J, Visited):-
  on_arrow(Model, I, New),
  (J =\= New->
    Visited /\ 2**New =:= 0,
    Visited2 :=  Visited \/ 2**New,
    can_go(Model, New, J, Visited2)
  ).

on_arrow(Model, St, New):-
  for(New, 0, 7),
    links(Model,St,New) =\= 0. % /\ (2**0 \/ 2**1)

[] :== []/S. 
[H|T] :== [H1|T1]/S :- 
  H:= H1/S,
  T:== T1/S.

[A2, B2] :== roll(Alfa, [A, B]) :- 
  A2 := A*cos(Alfa)+B*sin(Alfa),
  B2 := B*cos(Alfa)-A*sin(Alfa).

[] :== []+S*[]. 
[H|T] :== [H1|T1]+S*[H2|T2] :- 
  H:=H1+S*H2,
  T:==T1+S*T2.

