?-
  pen(1,0),
  G_Go=1, % 0 or 1
  G_Action=eye, % eye or point
  G_Point_Sensitivity=1, % from -4 to 4
  G_Eyes_Sensitivity=0, % from -4 to 4
  G_Time=0, % in seconds
  G_TimeStep=0.04, % 25 pic per sec
  G_Finger=0,
  G_X = 0,
  G_Y = 0,
  G_Z = -30,
  G_Zoom=300,
  G_A1 = 1, G_B1 = 0, G_C1 = 0,
  G_A2 = 0, G_B2 = 1, G_C2 = 0,
  G_A3 = 0, G_B3 = 0, G_C3 = 1,
  G_ScrollRange := 1000,
  G_ScrollPage :=100,
  window( size(700, 600), backgr(0xffff90), paint_indirectly, not(clean_before_paint),h_scroll,v_scroll).

win_func(init):-
  _ := set_timer(_, G_TimeStep, time_func),
  menu(normal, action(switch), text("Switch") ),
  menu(normal, action(move_down), text("<-") ),
  menu(normal, action(help2), text("5") ),
  menu(normal, action(move_up), text("->") ),
  menu(normal, action(leave_trace), text("Leave Trace") ),
  menu(normal, action(stop), text("Stop") ),
  menu(right, action(help), text("Help") ),
  set_scroll_range(_,vertical,0,G_ScrollRange),
  set_scroll_range(_,horizontal,0,G_ScrollRange),
  G_Scroll := (G_ScrollRange-G_ScrollPage)//2,
  set_scroll_page(_,vertical,G_ScrollPage),
  set_scroll_page(_,horizontal,G_ScrollPage),
  set_scroll_pos(_,vertical,G_Scroll),
  set_scroll_pos(_,horizontal,G_Scroll),
  show_text,
  change_sensitivity(new),
  show_go.

win_func(size(X, Y)):-
  G_X_size:=X/2,
  G_Y_size:=Y/2,
  update_window(_).

win_func(paint):-
  (G_Go=1, from_update_window->
    T is chronometer(),
    G_Time:=G_Time+T
  ),
  view_point(P),
  matrix(M),
  mat_mat(M,stretch(z, 1/G_Zoom),M2),
  vec_mat(P, stretch(-1), P1),
  vec_mat(P1, M2, P2),
  draw_figura(world([0,0,0],one), P2, M2).

win_func(mouse_click(X, Y)):-
  G_moveX:=X,
  G_moveY:=Y,
  set_capture(_).

win_func(mouse_click_up(X, Y)):-
  release_capture.

win_func(mouse_move(X, Y)):-
  (message_flags(left)->
    (G_Action=point->
      move_view_point(y, -2, (G_moveX-X)/G_X_size),
      move_view_point(x, -2, (G_moveY-Y)/G_Y_size)
    else
      roll_eyes(y, 0.2, (G_moveX-X)/G_X_size),
      roll_eyes(x, 0.2, (G_moveY-Y)/G_Y_size)
    ),
    G_moveX:=X,
    G_moveY:=Y
  ),
  show_text,
  update_window(_).

win_func(scroll(Direction,Type,Val)):-
  (Type=line->
    Step := 10
  else (Type=page->
    Step := 150
  else (Type=thumb_track->
    Step := 1,
    Old := Val,
    Val := Val - G_Scroll,
    G_Scroll := Old
  else % thumb_moved_to
    G_Scroll := (G_ScrollRange-G_ScrollPage)//2,
    fail
  ))),
  (Direction=horizontal-> % left / right
    (G_Action=point->
      move_view_point(y, 2, Step*Val/G_X_size)
    else
      roll_eyes(y, -0.2, Step*Val/G_X_size)
    )
  else (Direction=vertical-> % up / down
    (G_Action=point->
      move_view_point(x, 2, Step*Val/G_Y_size)
    else
      roll_eyes(x, -0.2, Step*Val/G_Y_size)
    )
  )),
  show_text,
  update_window(_).

win_func(mouse_wheel(Delta, _, _)):-
  (G_Action=point->
    move_view_point(z, 2, Delta)
  else (message_flags(left); message_flags(ctrl)->
    G_Zoom:=G_Zoom*1.01**(Delta*4**G_Eyes_Sensitivity)
  else
    roll_eyes(z, 0.2, Delta)
  )),
  show_text,
  update_window(_).

win_func(key_down(Key, Rep)):-
  (Key=12; Key=101->
    switch(press)
  else (49=<Key, Key=<57 ->
    change_sensitivity(Key-53)
  else
    action(G_Action, Key, Rep)
  )).

action(eye, Key, Repetition):-
  member(Key, [34, 99, 35, 97]), !,
  Rep is Repetition*4**G_Eyes_Sensitivity,
  (Key=34; Key=99->
    G_Zoom:=G_Zoom*1.01**Rep
  else
    G_Zoom:=G_Zoom/1.01**Rep
  ),
  show_text,
  update_window(_).

action(eye, Key, Repetition):-
  member(Key, [38, 104, 40, 98, 37, 100, 39, 102, 33, 105, 36, 103]), !,
  check_direction(Key, D),
  (member(Key, [38, 104, 37, 100, 33, 105])->
    roll_eyes(D, 0.01, Repetition)
  else
    roll_eyes(D, -0.01, Repetition)
  ),
  show_text,
  update_window(_).

action(point, Key, Repetition):-
  member(Key, [38, 104, 40, 98, 37, 100, 39, 102, 35, 97, 36, 103]), !,
  check_direction(Key, D),
  (member(Key, [38, 104, 37, 100, 35, 97])->
    move_view_point(D, -0.1, Repetition)
  else
    move_view_point(D,  0.1, Repetition)
  ),
  show_text,
  update_window(_).

action(_, _, _):-
  beep.

check_direction(Key, D):-
  (member(Key, [38, 104, 40, 98])->
    D=x
  else (member(Key, [37, 100, 39, 102])->
    D=y
  else
    D=z
  )).

move_view_point(D, S, Repetition):-
  view_point(P),
  matrix(M),
  transponirana(M, [V1, V2, V3]),
  (D=x->
    V=V2
  else (D=y->
    V=V1
  else
    V=V3
  )),
  Rep is S*Repetition*4**G_Point_Sensitivity,
  vec_mat(V, stretch(Rep), Vn),
  vec_p_vec(P, Vn, P2),
  change_view_point(P2).

roll_eyes(D, A, Repetition):-
  Rep is A*Repetition*4**G_Eyes_Sensitivity,
  matrix(M1),
  mat_mat(M1,rotation(D, Rep),M2),
  change_matrix(M2).

show_text:-
  (G_Action=point->
    set_text("3D Studio - View Point: ("+integer(G_X)+", "+integer(G_Y)+", "+integer(G_Z)+")", _)
  else
    matrix(M),
    vec_mat([0,0,1],M,[X,Y,Z]),
    set_text("3D Studio - Eyes Direction: ("+round(100*X)+", "+round(100*Y)+", "+round(100*Z)
      +") zoom: "+integer(G_Zoom), _)
  ).
change_sensitivity(New):-
  (G_Action=point->
    Which=G_Point_Sensitivity
  else
    Which=G_Eyes_Sensitivity
  ),
  (New=up->
    Which<4,
    Which:=Which+1
  else (New=down->
    Which> -4,
    Which:=Which-1
  else (New\=new->
    Which:=New
  ))),
  modify_menu(_, 3, _, print(5+Which)),
  draw_menu_bar(_).

show_go:-
  modify_menu(_, 6, _, (G_Go=0->"Go"else"Stop")),
  draw_menu_bar(_).

switch(press):-
  G_Action := (G_Action=eye -> point else eye),
  show_text,
  change_sensitivity(new).

move_down(press):-
  change_sensitivity(down).

move_up(press):-
  change_sensitivity(up).

stop(press):-
  chronometer(),
  G_Go:=1-G_Go,
  show_go.

leave_trace(press):-
  set_flags(_, change(paint_indirectly)),
  modify_menu(_, 5, _, (get_flag(_, paint_indirectly)->"Leave Trace" else "Without Trace")),
  draw_menu_bar(_),
  update_window(_).

help(press):-
  message("3D Studio",
  "Use the numpad arrows in order to move in the space, to roll eyes and to zoom.
With the key 5 you can switch between \"Eyes Direction\" and \"View Point\"

The second option is to use the mouse and its wheel in order to move and zoom.",i).

help2(press):-
  What is "the " + (G_Action=point-> "View Point" else "Eyes Direction")+" Sensitivity",
  message("Change " + What,
  "Use the arrows <- and -> in order to change " + What + ".

You can enter this sensitivity directly by the keys from 1 to 9.",i).

time_func(end):-
  G_Go=1,
  modyfy,
  update_window(_).

modyfy:- 
  X := 10*G_Time mod 60,
  (X < 30 ->
    G_Finger := 0.1+X/15
  else
    G_Finger := 0.1+30/15 -(X-30)/15
  ).

/***************/

draw_object([H|T], P, M):-
  draw_figura(H, P, M),
  draw_object(T, P, M).

calc_point(Vn, V1, P, M):-
  vec_mat(V1,M,Vt),
  vec_p_vec(Vt, P, Vn).

draw_figura(Figure(P1, M1),P, M):-
  vec_mat(P1,M,Pt),
  vec_p_vec(Pt, P, Pn),
  mat_mat(M1,M,M2),
  Figure(What, List),
  (What=shape->
    draw_line(List, Pn, M2)
  else (What=object->
    draw_object(List, Pn, M2)
  else
    draw_bezier_line(List, Pn, M2)
  )).
draw_figura(_,_,_).

draw_line([A | T], P, M):-
  calc_point(Pa, A, P, M),
  draw_line2(Pa, T, P, M).

draw_line2(Pa, [break | T], P, M):-
  !, draw_line(T, P, M).

draw_line2(Pa, [B | T], P, M):-
  calc_point(Pb, B, P, M),
  (calc_line(Pa, Pb, A1, B1, A2, B2)->
    line(A1, B1, A2, B2)
  ),
  draw_line2(Pb, T, P, M).

draw_bezier_line([A | T], P, M):-
  calc_point(Pa, A, P, M),
  draw_bezier_line2(Pa, T, P, M).

draw_bezier_line2(Pa, [B, C, D | T], P, M):-
  calc_point(Pb, B, P, M),
  calc_point(Pc, C, P, M),
  calc_point(Pd, D, P, M),
  (calc_line(Pa, Pb, A1, B1, A2, B2),
   calc_line(Pd, Pc, A4, B4, A3, B3)->
    %write((A2-A1)+","+(B2-B1)+";"+(A3-A4)+","+(B3-B4)),nl,
    bezier_line(A1, B1, A2, B2, A3, B3, A4, B4)
  ),
  draw_bezier_line2(Pd, T, P, M).

calc_line([X1, Y1, Z1], [X2, Y2, Z2], A1, B1, A2, B2):-
  (Z1>0->
    A1 is G_X_size+X1/Z1,
    B1 is G_Y_size+Y1/Z1,
    (Z2>0->
      A2 is G_X_size+X2/Z2,
      B2 is G_Y_size+Y2/Z2
    else
      far_point([X1, Y1, Z1], [X2, Y2, Z2], A1, B1, A2, B2)
    )
  else
    Z2>0,
    A2 is G_X_size+X2/Z2,
    B2 is G_Y_size+Y2/Z2,
    far_point([X2, Y2, Z2], [X1, Y1, Z1], A2, B2, A1, B1)
  ).

far_point([X1, Y1, Z1], [X2, Y2, Z2], A1, B1, A2, B2):-
  X is X1-(X1-X2)*Z1/(Z1-Z2), %(X2*Z1-X1*Z2)/(Z1-Z2)
  Y is Y1-(Y1-Y2)*Z1/(Z1-Z2),
  (abs(X)>abs(Y)->
    (X<0->
      A2 is -9*G_X_size,
      B2 is B1+(A2-A1)*Y/X
    else
      A2 is 11*G_X_size,
      B2 is B1+(A2-A1)*Y/X
    )
  else
    (Y<0->
      B2 is -9*G_Y_size,
      A2 is A1+(B2-B1)*X/Y
    else
      B2 is 11*G_Y_size,
      A2 is A1+(B2-B1)*X/Y
    )
  ).

/***************/

vec_p_vec([X1, Y1, Z1], [X2, Y2, Z2], [Xn, Yn, Zn]):-
  Xn is X1+X2,
  Yn is Y1+Y2,
  Zn is Z1+Z2.

vec_mat([X,Y,Z],[[A1,B1,C1],
                 [A2,B2,C2],
                 [A3,B3,C3]],[R1,R2,R3]):-
  R1 is X*A1+Y*A2+Z*A3,
  R2 is X*B1+Y*B2+Z*B3,
  R3 is X*C1+Y*C2+Z*C3.

vec_mat(V,one,V).

vec_mat([X,Y,Z],rotation(x,A),[X,Y2,Z2]):-
  Y2 is Y*cos(A)+Z*sin(A),
  Z2 is Z*cos(A)-Y*sin(A).

vec_mat([X,Y,Z],rotation(y,A),[X2,Y,Z2]):-
  X2 is X*cos(A)+Z*sin(A),
  Z2 is Z*cos(A)-X*sin(A).

vec_mat([X,Y,Z],rotation(z,A),[X2,Y2,Z]):-
  X2 is X*cos(A)+Y*sin(A),
  Y2 is Y*cos(A)-X*sin(A).

vec_mat([X,Y,Z],stretch(S),[X2,Y2,Z2]):-
  X2 is S*X,
  Y2 is S*Y,
  Z2 is S*Z.

vec_mat([X,Y,Z],stretch(x, S),[X2,Y,Z]):-
  X2 is S*X.

vec_mat([X,Y,Z],stretch(y, S),[X,Y2,Z]):-
  Y2 is S*Y.

vec_mat([X,Y,Z],stretch(z, S),[X,Y,Z2]):-
  Z2 is S*Z.

vec_mat(V,M1*M2,Vn):-
  vec_mat(V,M1,Vp),
  vec_mat(Vp,M2,Vn).

mat_mat(rotation(D,A),M,M2):-
  mat_mat([[1, 0, 0],
           [0, 1, 0],
           [0, 0, 1]],rotation(D,A), M1),
  mat_mat(M1,M,M2).

mat_mat(M1*M2,M,Mn):-
  mat_mat(M2,M,Mp),
  mat_mat(M1,Mp,Mn).

mat_mat(stretch(S),M,M2):-
  mat_mat(M,stretch(S),M2).

mat_mat(stretch(x, S),[V, V2, V3],[Vn, V2, V3]):-
  vec_mat(V,stretch(S),Vn).

mat_mat(stretch(y, S),[V1, V, V3],[V1, Vn, V3]):-
  vec_mat(V,stretch(S),Vn).

mat_mat(stretch(z, S),[V1, V2, V],[V1, V2, Vn]):-
  vec_mat(V,stretch(S),Vn).

mat_mat(one,M,M).

mat_mat([V1,V2,V3],M,[R1,R2,R3]):-
  vec_mat(V1,M,R1),
  vec_mat(V2,M,R2),
  vec_mat(V3,M,R3).

transponirana([[A1, B1, C1],
               [A2, B2, C2],
               [A3, B3, C3]], [[A1, A2, A3],
                               [B1, B2, B3],
                               [C1, C2, C3]]).

/***************/

matrix([[G_A1, G_B1, G_C1],
        [G_A2, G_B2, G_C2],
        [G_A3, G_B3, G_C3]]).

change_matrix([[A1, B1, C1],
               [A2, B2, C2],
               [A3, B3, C3]]):-
  G_A1:=A1, G_B1:=B1, G_C1:=C1,
  G_A2:=A2, G_B2:=B2, G_C2:=C2,
  G_A3:=A3, G_B3:=B3, G_C3:=C3.

view_point([G_X, G_Y, G_Z]).

change_view_point([X, Y, Z]):-
  G_X:=X, G_Y:=Y, G_Z:=Z.

