% Metro.pro, version 5.0
% from August 23, 2023

include("metro_map.h"). % try 1, 2, 3 with Start_from_the_beginning
include("metro_data.h").
include("metro_window.h").
include("metro_graphic.h").
include("metro_tools.h").

?-
  init_map,
  initialisation,
  (G_Start_from_the_beginning =:= 1 ->
    init_statistic,
    make_traffic(1),
    start_from_the_beginning
  else
    H is open("Heavy.metro", "rb"),  read_metro(H),  close(H)
  ),
  (G_BigSize=:=1 -> P:= 100, X := 1900, Y := 770 else P:= 10, X := 1350, Y := 700),
  window(title("Metro"), pos(0, P), size(X, Y), backgr(rgb(0, 200, 200)), paint_indirectly, h_scroll, v_scroll).

start_from_the_beginning:-
  Eights := G_Wagons_in_use//16,
  order_the_wagons_in_LR(0, 8*Eights),
  order_the_wagons_in_LR(2, 8*Eights+ G_Wagons_in_use mod 16).

order_the_wagons_in_LR(LR, All):-
  Fours := All//8,
  wJump(G_WagonNum) := -2,
  first(LR, 2) ::= G_WagonNum,
  first(LR, 6) ::= 4*Fours,
  order_the_wagons(LR, 2, Fours, 0),
  wJump(G_WagonNum) := -1,
  first(LR, 1) ::= G_WagonNum,
  first(LR, 5) ::= 4*Fours + All mod 8,
  order_the_wagons(LR, 1, Fours, All mod 8),
  make_ready_in(LR, 1).

order_the_wagons(LR, B, Fours, Remainder):-
  for(I, 0, 4*Fours+Remainder-1),
    create_wagon(LR, B, I),
  fail.
order_the_wagons(_, _, _, _).

create_wagon(LR, B, I):-
  (G_WagonNum mod 4 =:= 0 -> G_NextColor := (G_NextColor+1) mod 8),
  wagon_color(G_NextColor, Color),
  wColor(G_WagonNum) := Color,
  wPosoka(G_WagonNum) := 1-LR,
  StationX := finish_station_pos(LR-1),
  wagonX(G_WagonNum) := StationX + (LR-1)*(8000+(I+4.5)*wagon_size+G_ExtraDistance),
  wagonY(G_WagonNum) := ( (LR//2 + B) mod 2 =:= 0 -> 1000 else 1500),
  (I=:=0 ->
    wPrevious(G_WagonNum) := -1,
    wFlags(G_WagonNum) := 0
  else
    wNext(G_WagonNum-1) := G_WagonNum,
    wPrevious(G_WagonNum) := G_WagonNum-1,
    wFlags(G_WagonNum) := 1
  ),
  wNext(G_WagonNum) := -1,
  G_WagonNum := G_WagonNum + 1.
  

time_func(end):-
  G_Pause =:= 0,
  G_New_Map_Number =:= G_Map_Number,
  start_random_from(G_Time), % make the random deterministic
  (G_Time mod G_time_period =:= 0, G_Stop =:= 0 ->
    (trains_await(0)<10 -> trains_await(0) := trains_await(0) + 1),
    (trains_await(2)<10 -> trains_await(2) := trains_await(2) + 1)
  ),
  G_Time := G_Time + 1,
  % (G_Time mod 600 =:= 0 -> write("One minute for "+chronometer()/60), nl),
  move_wagons,
  (G_Time mod 10 =:= 0; G_FastSpeed =:= -1 ->
    set_main_window_text
  ),
  (G_Follow_the_wagon =:= 1 ->
    G_PosX := wagonX(G_The_Wagon) + G_PosX1 + G_PosX2,
    fix_small_pos(0.5*G_Xsize)
  ),
  G_Run_Invisible =:= 0,
  % (G_Time mod 2 =:= 0; G_FastSpeed =\= 1 -> % for slow computers
  update_window(_).

next_stop(0, 1, 1).
next_stop(1, 2, 2).
next_stop(2, 2, 1).
next_stop(3, 3, 3).
next_stop(4, 3, 2).
next_stop(5, 3, 1).
next_stop(6, 4, 4).
next_stop(7, 4, 3).
next_stop(8, 4, 2).
next_stop(9, 4, 1).

move_wagons:-
  for(I, 0, G_WagonNum-1),
  wOldX(I) := wagonX(I),
  (wCount(I)>0->
    wCount(I) := wCount(I)-1
  else (wAcceleration(I)>0->
    change_speed(I, 1)
  else (wAcceleration(I)<0->
    change_speed(I, -1)
  ))),
  (in_curve(I, X, Y, Sign, R)->
    Alfa := arccos((wagonX(I)-X)/R),
    %write(wagonX(I)+"="+Alfa+"="+cos(Alfa - wSpeed(I)/4625)*4625), nl,
    wagonX(I) := round( X+cos(Alfa - wSpeed(I)/R)*R ),
    wagonY(I) := round( Y+wPosoka(I)*Sign*(R-sin(Alfa - wSpeed(I)/R)*R) )
  else
    wagonX(I) := wagonX(I) + round(wSpeed(I))
  ),
  % (I=9, wSpeed(I)<200-> write("speed="+wSpeed(I)+" "+G_Time), nl),
  fail.
move_wagons:-
  for(I, 0, G_WagonNum-1),
  wFlags(I)/\1 =:= 0,
  NextSpeed := wSpeed(I) + (wPosoka(I)*wAcceleration(I) > 0 -> wAcceleration(I) else 0),
  (2*wPosoka(I)*(wStopAt(I)-wagonX(I)-wSpeed(I)-wAcceleration(I))<NextSpeed**2, wSpeed(I) =\= 0->
    (abs(wSpeed(I)) > G_Statistic_Max_Speed->
      G_Statistic_Max_Speed := abs(wSpeed(I))
    ),
    wFlags(I):=(wFlags(I)\/64)/\0xffdf, % 32
    Acceleration := -1*wPosoka(I), % 0.5*wSpeed(I)**2/(wagonX(I)-wStopAt(I)),
    move_train2(I, Acceleration, 0)
  else (wFlags(I)/\32=\=0->
    (wPrevious(I)=\= -1->
      calc_Distanse(Distanse, Limit, I),
      (2*Distanse<Limit, wSpeed(I) =\= 0->
        move_train2(I, -1*wPosoka(I), 0)
      else (Distanse<Limit, wSpeed(I) =\= 0->
        move_train2(I, -0.5*G_Crazy_Speed*wPosoka(I), 0)
      else (0.8*Distanse-100>Limit-> % 0.8 coef. and 100 coef.
        end_of_slowdown(I)
      else
        move_train2(I, 0, 0)
      )))
    else
      end_of_slowdown(I)
    )
  else (wFlags(I)/\4=:=0, wPrevious(I)=\= -1->
    calc_Distanse(Distanse, Limit, I),
    (Distanse<Limit->
      (abs(wSpeed(I)) > G_Statistic_Max_Speed->
        G_Statistic_Max_Speed := abs(wSpeed(I))
      ),
      wFlags(I):=wFlags(I)\/32,
      move_train2(I, -0.5*G_Crazy_Speed*wPosoka(I), 0)
    else (wFlags(I)/\64=\=0->
      wFlags(I):=wFlags(I)/\0xffbf,
      move_train2(I, -0.5*wPosoka(I), 0)
    ))
  else (wFlags(I)/\64=\=0->
    wFlags(I):=wFlags(I)/\0xffbf, % 64
    move_train2(I, -0.5*wPosoka(I), 0)
  )))),
  fail.
move_wagons:-
  for(I, 0, G_WagonNum-1),
  (wFlags(I)/\2=\=0 ->
    make_ready_2(I)
  else (wSpeed(I)=:=0, wCount(I)=:=0, wAcceleration(I)=:=0 ->
    (wFlags(I)/\8=\=0->
      make_ready_8(I)
    else (wFlags(I)/\4=\=0, (G_Accident =:= 0; wPosoka(I) =:= -1; abs(400000-wagonX(I))>4000) ->
      init_wagon(I)
    ))
  )),
  fail.
move_wagons:-
  for(I, 0, G_WagonNum-1),
  where_to_stop(Type, I, N, StationX),
  cross_point(Type, I, N, StationX),
  fail.
move_wagons.

change_speed(I, Sing):-
  (Sing*wTargetSpeed(I)-0.5 > Sing*(wSpeed(I)+wAcceleration(I))-> % 0.5 because too little speed does not move
    wSpeed(I) := wSpeed(I)+wAcceleration(I)
  else
    (abs(wTargetSpeed(I)-wSpeed(I))-0.5 > abs(wAcceleration(I))-> write("Opposite acceleration! i="+I+" speed=("+wSpeed(I)+","+wTargetSpeed(I)+") Acceleration="+wAcceleration(I)+" now="+G_Time), nl),
    wSpeed(I) := wTargetSpeed(I),
    wAcceleration(I) :=0,
    (wTargetSpeed(I)=:=0 ->
      (wFlags(I)/\8=:=0, abs(wStopAt(I)-wagonX(I))<5 -> % 5 coef.
        (wStop(I)=:=0, wFlags(I)/\2=:=0, wJump(I)>0->
          train_flag(I, 4),
          wCount(I):=wait_time+random(1+2*plus_minus)-plus_minus,
          (G_Unexpected_Delay=:=1, random(unexpected_delay_freq)=:=0 ->
            wCount(I):= wCount(I)+1+random(unexpected_delay_size), write(x)
          ),
          (wTime(I) =\= -1 ->
            G_Statistic_Stations := G_Statistic_Stations + 1,
            G_Statistic_Station_Stay := G_Statistic_Station_Stay + wCount(I)
          )
        )
      )      
    )
  ).

where_to_stop(Type, I, N, StationX):-
  station(yes, N, StationX, Semaphore, _, _),
  abs(StationX-wagonX(I))<43000,
  stop_point(Type, Point),
  (wPosoka(I)=:=1->
    wOldX(I) =< Point+StationX+wagon_size/2, Point+StationX+wagon_size/2 < wagonX(I)
  else
    wOldX(I) >= -1*Point+StationX-wagon_size/2, -1*Point+StationX-wagon_size/2 > wagonX(I)
  ).

stop_point(enter_switch, -5000).

cross_point(enter_switch, I, N, StationX):-
  (wFlags(I)/\1=:=0->
    wStop(I) := wStop(I)-1
  ),
  (wNext(I) =:= -1; wFlags(wNext(I))/\1 =:= 0 -> % switch
    N =:= finish_station(-1*wPosoka(I)),
    % then
    LR := 1-wPosoka(I),
    (wFlags(find_head(I)) /\ 16 =\= 0 ->
      busy(LR) := busy(LR) /\ (busy(LR)/\2 =\= 0 -> 0xfcd else 0xcfe) % (2,0x30 or 1,0x300)
    else
      wPrevious(wNext(I)) := -1,
      wNext(I) := -1,
      busy(LR) := busy(LR) /\ (busy(LR)/\2 =\= 0 -> 0xffd else 0xffe) % (2 or 1)
    )
  ).

